close all
clear
clc

syms f(x) g(x,y) h

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ONE DIMENSIONAL EXAMPLE %
%%%%%%%%%%%%%%%%%%%%%%%%%%%

f(x) = sin(x)*cos(x);

% second-order accurate second derivative
f_xx(x,h) = second_derivative(f(x),x,h);

% composition
% f'''(x) = f'(f''(x))
f_xxx(x,h) = first_derivative(f_xx(x,h),x,h);

% evaluation point
xval_f = pi;

% exact solution
f_E = diff(f,3);
f_exact = double(subs(f_E, xval_f));


%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TWO DIMENSIONAL EXAMPLE %
%%%%%%%%%%%%%%%%%%%%%%%%%%%

g(x,y) = sin(x)*cos(y) + cos(x)*sin(y);

% fourth-order accurate second derivative
gx2 = fourth_order_second_derivative(g(x,y),x,h);

% compositions
% g^(4,0)(x,y)  = g^(2,0)(g^(2,0)(x,y))
gx4 = fourth_order_second_derivative(gx2,x,h);

% g^(4,2)(x,y)  = g^(0,2)(g^(4,0)(x,y))
gx4y2 = fourth_order_second_derivative(gx4,y,h);

% g^(4,3)(x,y)  = g^(0,1)(g^(4,2)(x,y))
gx4y3 = fourth_order_first_derivative(gx4y2,x,h);

% exaluation points
xval_g = 2*pi;
yval_g = pi/3;

% exact solution
g_E = diff(g,x,x,x,x,y,y,y);
g_exact = double(subs(g_E,{x,y},{xval_g,yval_g}));

%%%%%%%%%%%%%%%%%%%%%
% CALCULATING ERROR %
%%%%%%%%%%%%%%%%%%%%%

% number of h values
hlen = 10;

% inialize arrays
hval = zeros(1,hlen+1);
error_f = zeros(1, hlen);
error_g = zeros(1, hlen);

% initial h value
hval(1) = 1/4;


for i=1:hlen
    
    % substitute h value in f_xxx and gx4y3
    fval = double(f_xxx(xval_f, hval(i)));
    gval = double(subs(gx4y3,{x,y,h},{xval_g,yval_g,hval(i)}));
    
    % calculate 2-norm error
    error_f(i) = norm((f_exact - fval),2);
    error_g(i) = norm((g_exact - gval),2);
    
    % divide h by 2 for next iteration
    hval(i+1) = hval(i)/2;
end

%%%%%%%%%%%%%%%%%
% LOG-LOG PLOTS %
%%%%%%%%%%%%%%%%%

clear h
h = hval(1:hlen);
ax = gca;

figure(1)
loglog(h,error_f,'-ksq','LineWidth',2);
grid;
set(gca,'LineWidth',1)
ylabel('error');
xlabel('h');
hold on
xref = 10.^(-2.5:.1:-0.5);
yref = exp(2*log(xref)+log(10^-1));
plot(xref,yref,'--k','LineWidth',2);
legend('||u_h- u||', 'h^2','Location','NORTHWEST','FontSize',12);
hold off;  

figure(2)
loglog(h,error_g,'-ksq','LineWidth',2);
grid;
set(gca,'LineWidth',1)
ylabel('error');
xlabel('h')
hold on
xref = 10.^(-2.5:.1:-0.5);
yref = exp(4*log(xref)+log(10^-4));
plot(xref,yref,'--k','LineWidth',2);
legend('||u_h- u||', 'h^4','Location','NORTHWEST','FontSize',12);
hold off; 

%%%%%%%%%%%%%%%%%%%%%
% CONVERGENCE RATES %
%%%%%%%%%%%%%%%%%%%%%

rate_f = zeros(1, hlen-1);
rate_g = zeros(1, hlen-1);
for i=1:hlen-1
    rate_f(i) = log2(error_f(i)/error_f(i+1));
    rate_g(i) = log2(error_g(i)/error_g(i+1));
end
convergence_rate_f = rate_f
convergence_rate_g = rate_g

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% COEFFICIENT PREDICTION 1D %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

p = polyfit(log(h), log(error_f), 1);
order_of_accuracy = round(p(1));
Ex = diff(f_E,x,order_of_accuracy);
f_coeff = double(subs(Ex,xval_f));
leading_term_coefficient_1D = exp(p(2))/f_coeff


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SECOND-ORDER ACCURATE DERIVATIVE FUNCTIONS %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function y = first_derivative(f,x,h)
% finite difference approximation for first derivative
% number of points: 3
% coefficients: -1/2 0 1/2 
% (generated using CoefficientGenerator mathematica file)

% f: function f(x)
% x: dimensional direction
% h: grid spacing in x-dir
% y: output function f'(x)

    y = (subs(f,x,x+h) - subs(f,x,x-h))/(2*h);
    
end

function y = second_derivative(f,x,h)
% finite difference approximation for second derivative
% number of points: 3
% coefficients: 1 -2 1
% (generated using CoefficientGenerator mathematica file)

% f: function f(x)
% x: dimensional direction
% h: grid spacing in x-dir
% y: output function f''(x)

    y = (subs(f,x,x+h) - 2*f + subs(f,x,x-h))/(h^2);
    
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FOURTH-ORDER ACCURATE DERIVATIVE FUNCTIONS %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function y = fourth_order_first_derivative(f,x,h)
% finite difference approximation for first derivative
% number of points: 5
% coefficients: 1/12 -2/3 0 2/3 -1/12
% (generated using CoefficientGenerator mathematica file)

% f: function f(x)
% x: dimensional direction
% h: grid spacing in x-dir
% y: output function f'(x)

    y = (-subs(f,x,x+2*h) + 8*subs(f,x,x+h) - 8*subs(f,x,x-h) ...
        + subs(f,x,x-2*h))/(12*h);
    
end

function y = fourth_order_second_derivative(f,x,h)
% finite difference approximation for second derivative
% number of points: 5
% coefficients: -1/12 4/3 -5/2 4/3 -1/12
% (generated using CoefficientGenerator mathematica file)

% f: function f(x)
% x: dimensional direction
% h: grid spacing in x-dir
% y: output function f''(x)

    y = (-subs(f,x,x+2*h) + 16*subs(f,x,x+h) -30*f + 16*subs(f,x,x-h) ...
        - subs(f,x,x-2*h))/(12*h^2);
    
end